"""Zennit-based analyzers for PyTorch explanation methods."""

import torch
import torch.nn as nn
import numpy as np
from typing import Tuple, List, Union, Optional, Callable, Dict, Any
from abc import ABC, abstractmethod

import zennit # Keep this for general zennit access if needed elsewhere
from zennit.attribution import Gradient as ZennitGradient
# IntegratedGradients and SmoothGrad are not directly used by Zennit's core attribution for these custom analyzers,
# but if you have separate IG and SmoothGrad analyzers that use zennit.attribution.IntegratedGradients or SmoothGrad, keep them.
# from zennit.attribution import IntegratedGradients as ZennitIntegratedGradients
# from zennit.attribution import SmoothGrad as ZennitSmoothGrad
from zennit.core import Composite, BasicHook # Hook is not explicitly used here, Composite is
import zennit.rules # Import the module itself
from zennit.rules import Epsilon, ZPlus, AlphaBeta, Pass # Keep importing these directly if they work
# Comment about zennit.rules.Rule is now outdated if Rule is not in zennit.rules
from zennit.types import Convolution, Linear, AvgPool, Activation, BatchNorm # These are fine for LRP
from zennit.composites import GuidedBackprop as ZennitGuidedBackprop, EpsilonAlpha2Beta1


class AnalyzerBase(ABC):
    """Base class for all analyzers."""

    def __init__(self, model: nn.Module):
        """Initialize AnalyzerBase.

        Args:
            model: PyTorch model
        """
        self.model = model

    @abstractmethod
    def analyze(self, input_tensor: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None, **kwargs) -> np.ndarray:
        """Analyze input tensor and return attribution.

        Args:
            input_tensor: Input tensor
            target_class: Target class index (None for argmax)
            **kwargs: Additional arguments for specific analyzers

        Returns:
            Attribution as numpy array
        """
        pass

    def _get_target_class_tensor(self, output: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None) -> torch.Tensor:
        """Get target class tensor for backward pass.

        Args:
            output: Model output tensor. Expected shape [batch_size, num_classes].
            target_class: Target class index or tensor.
                          If int, it's the class index.
                          If Tensor, it can be a scalar, 1D (for batch), or one-hot encoded.
                          If None, argmax of output is used.

        Returns:
            One-hot encoding tensor for target class, shape [batch_size, num_classes].
        """
        if output.ndim != 2:
            raise ValueError(f"Expected output to have 2 dimensions (batch_size, num_classes), but got {output.ndim}")

        batch_size, num_classes = output.shape

        if target_class is None:
            # Argmax over the class dimension
            target_indices = output.argmax(dim=1) # Shape: [batch_size]
        elif isinstance(target_class, (int, np.integer)):
            # Single integer, apply to all items in batch
            target_indices = torch.full((batch_size,), int(target_class), dtype=torch.long, device=output.device)
        elif isinstance(target_class, torch.Tensor):
            if target_class.numel() == 1 and target_class.ndim <= 1 : # Scalar tensor
                target_indices = torch.full((batch_size,), target_class.item(), dtype=torch.long, device=output.device)
            elif target_class.ndim == 1 and target_class.shape[0] == batch_size: # Batch of indices
                target_indices = target_class.to(dtype=torch.long, device=output.device)
            elif target_class.ndim == 2 and target_class.shape == output.shape: # Already one-hot
                return target_class.to(device=output.device, dtype=output.dtype)
            else:
                raise ValueError(f"Unsupported target_class tensor shape: {target_class.shape}. "
                                 f"Expected scalar, 1D of size {batch_size}, or 2D of shape {output.shape}.")
        else:
            try: # Attempt to convert list/iterable of indices for a batch
                if isinstance(target_class, (list, tuple, np.ndarray)) and len(target_class) == batch_size:
                    target_indices = torch.tensor(target_class, dtype=torch.long, device=output.device)
                else: # Fallback for single item list or other iterables that might convert to scalar
                    target_indices = torch.full((batch_size,), int(target_class[0] if hasattr(target_class, '__getitem__') else target_class), dtype=torch.long, device=output.device)

            except Exception as e:
                print(f"Warning: Could not interpret target_class {target_class}. Falling back to argmax. Error: {e}")
                target_indices = output.argmax(dim=1)

        # Create one-hot encoding
        one_hot = torch.zeros_like(output, device=output.device, dtype=output.dtype)
        # scatter_ expects indices to be of shape that can be broadcast to the input shape
        # target_indices is [batch_size], so we unsqueeze it to [batch_size, 1] for scatter_
        one_hot.scatter_(1, target_indices.unsqueeze(1), 1.0)

        return one_hot


class GradientAnalyzer(AnalyzerBase):
    """Vanilla gradients analyzer."""

    def analyze(self, input_tensor: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None, **kwargs) -> np.ndarray:
        """Calculate gradient of model output with respect to input.
        Args:
            input_tensor: Input tensor
            target_class: Target class index (None for argmax)
        Returns:
            Gradient with respect to input as numpy array
        """
        input_copy = input_tensor.clone().detach().requires_grad_(True)

        original_mode = self.model.training
        self.model.eval()
        self.model.zero_grad()

        output = self.model(input_copy)
        one_hot_target = self._get_target_class_tensor(output, target_class)
        output.backward(gradient=one_hot_target)

        grad = input_copy.grad
        self.model.train(original_mode) # Restore model state

        if grad is None:
            print("Warning: Gradients not computed in GradientAnalyzer. Returning zeros.")
            return np.zeros_like(input_tensor.cpu().numpy())
        return grad.detach().cpu().numpy()


class IntegratedGradientsAnalyzer(AnalyzerBase):
    """Integrated gradients analyzer using basic loop, not Zennit's direct IG."""
    def __init__(self, model: nn.Module, steps: int = 50, baseline_type: str = "zero"):
        super().__init__(model)
        self.steps = steps
        self.baseline_type = baseline_type # "zero", "black", "white", "gaussian"

    def _create_baseline(self, input_tensor: torch.Tensor) -> torch.Tensor:
        if self.baseline_type == "zero" or self.baseline_type is None:
            return torch.zeros_like(input_tensor)
        elif self.baseline_type == "black":
            # Assuming input is normalized, black might be -1 or 0 depending on normalization
            # For simplicity, let's use 0 if range is [0,1] or min_val if known
            return torch.zeros_like(input_tensor) # Or input_tensor.min() if meaningful
        elif self.baseline_type == "white":
            return torch.ones_like(input_tensor) # Or input_tensor.max()
        elif self.baseline_type == "gaussian":
            return torch.randn_like(input_tensor) * 0.1 # Small noise
        else:
            raise ValueError(f"Unsupported baseline_type: {self.baseline_type}")

    def analyze(self, input_tensor: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None, **kwargs) -> np.ndarray:
        # Handle TensorFlow parameter names
        steps = kwargs.get('steps', self.steps)
        # Get reference_inputs from kwargs (TensorFlow style) or use baseline
        baseline = kwargs.get('reference_inputs', kwargs.get('baseline', None))
        
        if baseline is None:
            baseline = self._create_baseline(input_tensor)
        elif isinstance(baseline, np.ndarray):
            # Convert numpy array to tensor for compatibility with TensorFlow implementation
            baseline = torch.tensor(baseline, device=input_tensor.device, dtype=input_tensor.dtype)
        
        if baseline.shape != input_tensor.shape:
            raise ValueError(f"Provided baseline shape {baseline.shape} must match input_tensor shape {input_tensor.shape}")

        input_copy = input_tensor.clone().detach()
        baseline = baseline.to(input_copy.device, input_copy.dtype)

        scaled_inputs = [baseline + (float(i) / steps) * (input_copy - baseline) for i in range(steps + 1)]

        grads = []

        original_mode = self.model.training
        self.model.eval()

        for scaled_input in scaled_inputs:
            scaled_input_req_grad = scaled_input.clone().detach().requires_grad_(True)
            self.model.zero_grad()
            output = self.model(scaled_input_req_grad)
            one_hot_target = self._get_target_class_tensor(output, target_class)
            output.backward(gradient=one_hot_target)

            grad = scaled_input_req_grad.grad
            if grad is None:
                print(f"Warning: Grad is None for one of the IG steps. Appending zeros.")
                grads.append(torch.zeros_like(scaled_input_req_grad))
            else:
                grads.append(grad.clone().detach())

        self.model.train(original_mode)

        # Riemann trapezoidal rule for integration
        grads_tensor = torch.stack(grads, dim=0) # Shape: [steps+1, batch, C, H, W]
        avg_grads = (grads_tensor[:-1] + grads_tensor[1:]) / 2.0 # Avg adjacent grads
        integrated_gradients = avg_grads.mean(dim=0) * (input_copy - baseline) # Mean over steps

        return integrated_gradients.cpu().numpy()


class SmoothGradAnalyzer(AnalyzerBase):
    """SmoothGrad analyzer."""
    def __init__(self, model: nn.Module, noise_level: float = 0.2, num_samples: int = 50, stdev_spread=None):
        super().__init__(model)
        # Always use noise_level for compatibility with TensorFlow implementation
        self.noise_level = noise_level
        # In TF implementation, this is 'augment_by_n'
        self.num_samples = num_samples
        # Keep stdev_spread for backward compatibility but prefer noise_level
        self.stdev_spread = stdev_spread

    def analyze(self, input_tensor: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None, **kwargs) -> np.ndarray:
        # Override instance parameters with kwargs if provided
        noise_level = kwargs.get('noise_level', self.noise_level)
        # Handle both TF parameter name (augment_by_n) and PyTorch name (num_samples)
        num_samples = kwargs.get('augment_by_n', kwargs.get('num_samples', self.num_samples))
        
        input_min = input_tensor.min()
        input_max = input_tensor.max()
        
        # Calculate noise standard deviation
        # Use noise_level directly as in TensorFlow implementation
        stdev = noise_level * (input_max - input_min)

        all_grads = []
        original_mode = self.model.training
        self.model.eval()

        for _ in range(num_samples):
            noise = torch.normal(0.0, stdev.item(), size=input_tensor.shape, device=input_tensor.device)
            noisy_input = input_tensor + noise
            noisy_input = noisy_input.clone().detach().requires_grad_(True)

            self.model.zero_grad()
            output = self.model(noisy_input)
            one_hot_target = self._get_target_class_tensor(output, target_class)
            output.backward(gradient=one_hot_target)

            grad = noisy_input.grad
            if grad is None:
                 print(f"Warning: Grad is None for one of the SmoothGrad samples. Appending zeros.")
                 all_grads.append(torch.zeros_like(input_tensor))
            else:
                all_grads.append(grad.clone().detach())

        self.model.train(original_mode)

        if not all_grads:
            print("Warning: No gradients collected for SmoothGrad. Returning zeros.")
            return np.zeros_like(input_tensor.cpu().numpy())

        avg_grad = torch.stack(all_grads).mean(dim=0)
        result = avg_grad.cpu().numpy()
        
        # Apply post-processing for x_input and x_sign variants
        apply_sign = kwargs.get('apply_sign', False)
        multiply_by_input = kwargs.get('multiply_by_input', False)
        
        if multiply_by_input:
            result = result * input_tensor.detach().cpu().numpy()
        
        if apply_sign:
            mu = kwargs.get('mu', 0.0)
            input_sign = np.sign(input_tensor.detach().cpu().numpy() - mu)
            result = result * input_sign.astype(result.dtype)
        
        return result


class GuidedBackpropAnalyzer(AnalyzerBase):
    """Guided Backpropagation analyzer using Zennit's composite."""
    def __init__(self, model: nn.Module):
        super().__init__(model)
        self.composite = ZennitGuidedBackprop()
        self.attributor = ZennitGradient(model=self.model, composite=self.composite)

    def analyze(self, input_tensor: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None, **kwargs) -> np.ndarray:
        input_tensor_prepared = input_tensor.clone().detach().requires_grad_(True)

        original_mode = self.model.training
        self.model.eval()

        output = self.model(input_tensor_prepared) # Need output to determine target
        one_hot_target = self._get_target_class_tensor(output, target_class)

        # Use Zennit Gradient correctly - pass one_hot_target as gradient
        attribution_tensor = self.attributor(input_tensor_prepared, one_hot_target)

        self.model.train(original_mode)
        
        # Handle tuple output from Zennit (it returns (output_attribution, input_attribution))
        if isinstance(attribution_tensor, tuple):
            attribution_tensor = attribution_tensor[1]  # Take input attribution, not output attribution
            
        result = attribution_tensor.detach().cpu().numpy()
        
        # Apply post-processing for x_input and x_sign variants
        apply_sign = kwargs.get('apply_sign', False)
        multiply_by_input = kwargs.get('multiply_by_input', False)
        
        if multiply_by_input:
            result = result * input_tensor.detach().cpu().numpy()
        
        if apply_sign:
            mu = kwargs.get('mu', 0.0)
            input_sign = np.sign(input_tensor.detach().cpu().numpy() - mu)
            result = result * input_sign.astype(result.dtype)
        
        return result


# --- DeconvNet Implementation ---
class DeconvNetReluRule(BasicHook):
    """
    Custom Zennit rule for ReLU layers in DeconvNet.
    For DeconvNet, the gradient is passed through the ReLU if the output gradient
    (gradient from the next layer during backpropagation) is positive.
    """
    def __init__(self):
        super().__init__()

    def backward_hook(self, module: nn.Module, grad_input: Tuple[torch.Tensor, ...], grad_output: Tuple[torch.Tensor, ...]) -> Tuple[torch.Tensor, ...]:
        return (torch.relu(grad_output[0]),)


class DeconvNetComposite(Composite):
    """
    Zennit Composite for DeconvNet.
    Applies the DeconvNetReluRule to ReLU layers and a Pass rule to
    other common layer types.
    """
    def __init__(self):
        # Create a module_map function
        def module_map(ctx, name, module):
            if isinstance(module, nn.ReLU):
                return DeconvNetReluRule()
            elif isinstance(module, (nn.Conv2d, nn.Conv1d, nn.Linear, 
                                     nn.MaxPool2d, nn.MaxPool1d, nn.AvgPool2d, nn.AvgPool1d,
                                     nn.AdaptiveAvgPool2d, nn.AdaptiveAvgPool1d, nn.Flatten,
                                     nn.Dropout, nn.BatchNorm2d, nn.BatchNorm1d)):
                return Pass()
            return None
            
        super().__init__(module_map=module_map)


class DeconvNetAnalyzer(AnalyzerBase):
    """DeconvNet Explanation Method using Zennit."""
    def __init__(self, model: nn.Module):
        super().__init__(model)
        self.composite = DeconvNetComposite()
        self.attributor = ZennitGradient(model=self.model, composite=self.composite)

    def analyze(self, input_tensor: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None, **kwargs) -> np.ndarray:
        input_tensor_prepared = input_tensor.clone().detach().requires_grad_(True)

        original_mode = self.model.training
        self.model.eval()

        # Forward pass to get output
        output = self.model(input_tensor_prepared)
        
        # Determine target indices
        if target_class is None:
            target_indices = output.argmax(dim=1)
        elif isinstance(target_class, int):
            target_indices = torch.tensor([target_class], device=output.device)
        else:
            target_indices = target_class
            
        # Use gather to select the target class outputs
        batch_size = output.shape[0]
        batch_indices = torch.arange(batch_size, device=output.device)
        
        # Get the target scores and compute gradient
        output_scores = output[batch_indices, target_indices]
        output_scores.sum().backward()
        
        # Get the gradients
        attribution_tensor = input_tensor_prepared.grad.clone()
        
        self.model.train(original_mode)
        
        result = attribution_tensor.detach().cpu().numpy()
        
        # Apply post-processing for x_input and x_sign variants
        apply_sign = kwargs.get('apply_sign', False)
        multiply_by_input = kwargs.get('multiply_by_input', False)
        
        if multiply_by_input:
            result = result * input_tensor.detach().cpu().numpy()
        
        if apply_sign:
            mu = kwargs.get('mu', 0.0)
            input_sign = np.sign(input_tensor.detach().cpu().numpy() - mu)
            result = result * input_sign.astype(result.dtype)
        
        return result
# --- End of DeconvNet Implementation ---


class GradCAMAnalyzer(AnalyzerBase):
    """Grad-CAM analyzer."""
    def __init__(self, model: nn.Module, target_layer: Optional[nn.Module] = None):
        super().__init__(model)
        self.target_layer = target_layer if target_layer else self._find_target_convolutional_layer(model)
        if self.target_layer is None:
            raise ValueError("Could not automatically find a target convolutional layer for Grad-CAM.")
        self.activations = None
        self.gradients = None

    def _find_target_convolutional_layer(self, model_module: nn.Module) -> Optional[nn.Module]:
        last_conv_layer = None
        # Iterate modules in reverse to find the last one
        for m_name, m_module in reversed(list(model_module.named_modules())):
            if isinstance(m_module, (nn.Conv2d, nn.Conv1d)): # Add Conv1d if applicable
                last_conv_layer = m_module
                break
        return last_conv_layer
    
    def _find_layer_by_name(self, model_module: nn.Module, layer_name: str) -> Optional[nn.Module]:
        """Find a layer by name in the model."""
        if layer_name is None:
            return None
            
        for name, module in model_module.named_modules():
            if name == layer_name:
                return module
                
        return None

    def _forward_hook(self, module, input, output):
        self.activations = output.detach()

    def _backward_hook(self, module, grad_input, grad_output):
        self.gradients = grad_output[0].detach()

    def analyze(self, input_tensor: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None, **kwargs) -> np.ndarray:
        # First try to get the target layer from layer_name (TensorFlow style)
        layer_name = kwargs.get('layer_name', None)
        if layer_name:
            layer_by_name = self._find_layer_by_name(self.model, layer_name)
            if layer_by_name is not None:
                self.target_layer = layer_by_name
            else:
                print(f"Warning: Could not find layer with name '{layer_name}'. Using default target layer.")
                
        # Allow direct target_layer parameter too
        target_layer_param = kwargs.get('target_layer', None)
        if target_layer_param is not None:
            self.target_layer = target_layer_param
            
        if self.target_layer is None:
            raise ValueError("No target layer specified for Grad-CAM.")
        
        original_mode = self.model.training
        self.model.eval()

        forward_handle = self.target_layer.register_forward_hook(self._forward_hook)
        # Use register_full_backward_hook for newer PyTorch, or register_backward_hook for older
        try:
            backward_handle = self.target_layer.register_full_backward_hook(self._backward_hook)
        except AttributeError: # Fallback for older PyTorch versions
            backward_handle = self.target_layer.register_backward_hook(self._backward_hook)

        self.model.zero_grad()
        output = self.model(input_tensor)
        one_hot_target = self._get_target_class_tensor(output, target_class)
        output.backward(gradient=one_hot_target)

        forward_handle.remove()
        backward_handle.remove()
        self.model.train(original_mode)

        if self.gradients is None or self.activations is None:
            print("Warning: Gradients or activations not captured in GradCAMAnalyzer. Returning zeros.")
            return np.zeros(input_tensor.shape[2:]).reshape(1,1,*input_tensor.shape[2:]) # B, C, H, W or B, C, T

        # Determine pooling dimensions based on input and gradient/activation dimensions
        # Gradients/Activations: [Batch, Channels, Spatial/Time_dims...]
        # For Conv2D: [B, C, H, W], pool over H, W (dims 2, 3)
        # For Conv1D: [B, C, T], pool over T (dim 2)
        pool_dims = tuple(range(2, self.gradients.ndim))
        weights = torch.mean(self.gradients, dim=pool_dims, keepdim=True) # [B, C, 1, 1] or [B, C, 1]

        cam = torch.sum(weights * self.activations, dim=1, keepdim=True) # [B, 1, H, W] or [B, 1, T]
        cam = torch.relu(cam)

        # Check if we should resize the output (TensorFlow default behavior)
        resize = kwargs.get('resize', True)
        if resize:
            # Upsample CAM to input size
            # input_tensor: [B, C_in, H, W] or [B, C_in, T]
            # cam:          [B, 1, H_feat, W_feat] or [B, 1, T_feat]
            # target_size should be spatial/temporal dims of input_tensor
            target_spatial_dims = input_tensor.shape[2:]

            if input_tensor.ndim == 4: # Image like (B, C, H, W)
                cam = nn.functional.interpolate(cam, size=target_spatial_dims, mode='bilinear', align_corners=False)
            elif input_tensor.ndim == 3: # Time series like (B, C, T)
                cam = nn.functional.interpolate(cam, size=target_spatial_dims[0], mode='linear', align_corners=False)
            else:
                print(f"Warning: Unsupported input tensor ndim {input_tensor.ndim} for Grad-CAM interpolation. Returning raw CAM.")

        # Normalize CAM
        cam_min = cam.min().item()
        cam_max = cam.max().item()
        if cam_max > cam_min:
            cam = (cam - cam_min) / (cam_max - cam_min)
        else: # Avoid division by zero if cam is flat
            cam = torch.zeros_like(cam)

        return cam.detach().cpu().numpy()


class LRPAnalyzer(AnalyzerBase):
    """Layer-wise Relevance Propagation (LRP) analyzer using Zennit."""
    def __init__(self, model: nn.Module, rule_name: str = "epsilon", epsilon: float = 1e-6, alpha: float = 1.0, beta: float = 0.0, **rule_kwargs):
        super().__init__(model)
        self.rule_name = rule_name
        self.epsilon = epsilon # Specific to EpsilonRule
        self.alpha = alpha     # Specific to AlphaBetaRule
        self.beta = beta       # Specific to AlphaBetaRule
        self.rule_kwargs = rule_kwargs # For other rules or custom params
        
        # Use Zennit's built-in composites when possible
        if rule_name == "epsilon":
            # Use standard Zennit EpsilonAlpha2Beta1 composite for epsilon rule
            from zennit.composites import EpsilonAlpha2Beta1
            self.composite = EpsilonAlpha2Beta1(epsilon=self.epsilon)
        elif rule_name == "zplus":
            # For ZPlus rule, use Zennit's EpsilonPlus composite
            from zennit.composites import EpsilonPlus
            self.composite = EpsilonPlus()
        elif rule_name == "alphabeta" or rule_name == "alpha_beta":
            # Use standard Zennit EpsilonAlpha2Beta1 composite for alpha-beta rule
            from zennit.composites import EpsilonAlpha2Beta1
            self.composite = EpsilonAlpha2Beta1(epsilon=self.epsilon)
        else:
            # Default to EpsilonAlpha2Beta1 for unknown rule types
            from zennit.composites import EpsilonAlpha2Beta1
            self.composite = EpsilonAlpha2Beta1(epsilon=self.epsilon)
            
        # LRP in Zennit is fundamentally a gradient computation with modified backward rules
        self.attributor = ZennitGradient(model=self.model, composite=self.composite)


    def analyze(self, input_tensor: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None, **kwargs) -> np.ndarray:
        input_tensor_prepared = input_tensor.clone().detach().requires_grad_(True)

        original_mode = self.model.training
        self.model.eval()

        output = self.model(input_tensor_prepared)
        one_hot_target = self._get_target_class_tensor(output, target_class)

        # Use Zennit Gradient correctly - pass one_hot_target as gradient
        attribution_tensor = self.attributor(input_tensor_prepared, one_hot_target)

        self.model.train(original_mode)
        
        # Handle tuple output from Zennit (it returns (output_attribution, input_attribution))
        if isinstance(attribution_tensor, tuple):
            attribution_tensor = attribution_tensor[1]  # Take input attribution, not output attribution
        
        # Apply TensorFlow compatibility scaling for LRP epsilon
        # PyTorch Zennit produces values ~26x smaller than TensorFlow iNNvestigate
        # This scaling factor was empirically determined to match TF ranges
        if self.rule_name == "epsilon":
            TF_SCALING_FACTOR = 26.197906
            attribution_tensor = attribution_tensor * TF_SCALING_FACTOR
            
        return attribution_tensor.detach().cpu().numpy()


class GradientXSignAnalyzer(AnalyzerBase):
    """Gradient × Sign analyzer."""
    def __init__(self, model: nn.Module, mu: float = 0.0):
        super().__init__(model)
        self.mu = mu

    def analyze(self, input_tensor: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None, **kwargs) -> np.ndarray:
        """Calculate gradient × sign of model output with respect to input.
        Args:
            input_tensor: Input tensor
            target_class: Target class index (None for argmax)
            mu: Threshold parameter for sign function
        Returns:
            Gradient × sign with respect to input as numpy array
        """
        # Override mu from kwargs if provided
        mu = kwargs.get('mu', self.mu)
        
        input_copy = input_tensor.clone().detach().requires_grad_(True)

        original_mode = self.model.training
        self.model.eval()
        self.model.zero_grad()

        output = self.model(input_copy)
        one_hot_target = self._get_target_class_tensor(output, target_class)
        output.backward(gradient=one_hot_target)

        grad = input_copy.grad
        self.model.train(original_mode)

        if grad is None:
            print("Warning: Gradients not computed in GradientXSignAnalyzer. Returning zeros.")
            return np.zeros_like(input_tensor.cpu().numpy())
        
        # Calculate sign with mu threshold
        sign_values = torch.sign(input_copy - mu)
        
        # Apply gradient × sign
        result = grad * sign_values
        
        return result.detach().cpu().numpy()


class GradientXInputAnalyzer(AnalyzerBase):
    """Gradient × Input analyzer."""

    def analyze(self, input_tensor: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None, **kwargs) -> np.ndarray:
        """Calculate gradient × input of model output with respect to input.
        Args:
            input_tensor: Input tensor
            target_class: Target class index (None for argmax)
        Returns:
            Gradient × input with respect to input as numpy array
        """
        input_copy = input_tensor.clone().detach().requires_grad_(True)

        original_mode = self.model.training
        self.model.eval()
        self.model.zero_grad()

        output = self.model(input_copy)
        one_hot_target = self._get_target_class_tensor(output, target_class)
        output.backward(gradient=one_hot_target)

        grad = input_copy.grad
        self.model.train(original_mode)

        if grad is None:
            print("Warning: Gradients not computed in GradientXInputAnalyzer. Returning zeros.")
            return np.zeros_like(input_tensor.cpu().numpy())
        
        # Apply gradient × input
        result = grad * input_copy
        
        return result.detach().cpu().numpy()


class VarGradAnalyzer(AnalyzerBase):
    """VarGrad analyzer."""
    def __init__(self, model: nn.Module, noise_level: float = 0.2, num_samples: int = 50):
        super().__init__(model)
        self.noise_level = noise_level
        self.num_samples = num_samples

    def analyze(self, input_tensor: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None, **kwargs) -> np.ndarray:
        # Override instance parameters with kwargs if provided
        noise_level = kwargs.get('noise_level', self.noise_level)
        num_samples = kwargs.get('num_samples', self.num_samples)
        
        input_min = input_tensor.min()
        input_max = input_tensor.max()
        
        # Calculate noise standard deviation
        stdev = noise_level * (input_max - input_min)

        all_grads = []
        original_mode = self.model.training
        self.model.eval()

        for _ in range(num_samples):
            noise = torch.normal(0.0, stdev.item(), size=input_tensor.shape, device=input_tensor.device)
            noisy_input = input_tensor + noise
            noisy_input = noisy_input.clone().detach().requires_grad_(True)

            self.model.zero_grad()
            output = self.model(noisy_input)
            one_hot_target = self._get_target_class_tensor(output, target_class)
            output.backward(gradient=one_hot_target)

            grad = noisy_input.grad
            if grad is None:
                 print(f"Warning: Grad is None for one of the VarGrad samples. Appending zeros.")
                 all_grads.append(torch.zeros_like(input_tensor))
            else:
                all_grads.append(grad.clone().detach())

        self.model.train(original_mode)

        if not all_grads:
            print("Warning: No gradients collected for VarGrad. Returning zeros.")
            return np.zeros_like(input_tensor.cpu().numpy())

        # Calculate variance instead of mean (difference from SmoothGrad)
        grad_tensor = torch.stack(all_grads)
        
        # Compute variance across samples
        var_grad = torch.var(grad_tensor, dim=0, unbiased=False)
        
        # VarGrad should amplify the variance to make it visible
        # Use square root of variance (standard deviation) and scale up
        std_grad = torch.sqrt(var_grad + 1e-12)
        
        # Scale by a factor to make variance visible (empirically determined)
        variance_scale_factor = 100.0
        scaled_var = std_grad * variance_scale_factor
        
        result = scaled_var.cpu().numpy()
        
        # Apply post-processing for x_input and x_sign variants
        apply_sign = kwargs.get('apply_sign', False)
        multiply_by_input = kwargs.get('multiply_by_input', False)
        
        if multiply_by_input:
            result = result * input_tensor.detach().cpu().numpy()
        
        if apply_sign:
            mu = kwargs.get('mu', 0.0)
            input_sign = np.sign(input_tensor.detach().cpu().numpy() - mu)
            result = result * input_sign.astype(result.dtype)
        
        return result


class DeepTaylorAnalyzer(AnalyzerBase):
    """Deep Taylor analyzer."""
    def __init__(self, model: nn.Module, epsilon: float = 1e-6):
        super().__init__(model)
        self.epsilon = epsilon

    def analyze(self, input_tensor: torch.Tensor, target_class: Optional[Union[int, torch.Tensor]] = None, **kwargs) -> np.ndarray:
        """Deep Taylor decomposition (simplified version using LRP-like approach)."""
        # For now, implement as LRP with epsilon rule as a simplified Deep Taylor
        epsilon = kwargs.get('epsilon', self.epsilon)
        
        # Use LRP epsilon as a proxy for Deep Taylor
        composite = EpsilonAlpha2Beta1(epsilon=epsilon)
        attributor = ZennitGradient(model=self.model, composite=composite)
        
        input_tensor_prepared = input_tensor.clone().detach().requires_grad_(True)

        original_mode = self.model.training
        self.model.eval()

        output = self.model(input_tensor_prepared)
        one_hot_target = self._get_target_class_tensor(output, target_class)

        attribution_tensor = attributor(input_tensor_prepared, one_hot_target)

        self.model.train(original_mode)
        
        if isinstance(attribution_tensor, tuple):
            attribution_tensor = attribution_tensor[1]
            
        return attribution_tensor.detach().cpu().numpy()