"""
Custom Zennit hooks that exactly match iNNvestigate's mathematical formulations.
These hooks implement the exact same operations as iNNvestigate's FlatRule and WSquareRule.

This fixes the correlation issues by ensuring PyTorch Zennit produces identical
mathematical results to TensorFlow iNNvestigate.
"""

import torch
import torch.nn as nn
from zennit.core import Hook, Stabilizer, Composite
from zennit.types import Convolution, Linear, BatchNorm, Activation, AvgPool
from typing import Optional, Union


class InnvestigateFlatHook(Hook):
    """
    Custom Flat hook that exactly matches iNNvestigate's FlatRule implementation.
    
    From iNNvestigate: FlatRule sets all weights to ones and no biases,
    then uses SafeDivide operations for relevance redistribution.
    
    CRITICAL FIX: Handles numerical instability when flat outputs are near zero.
    """
    
    def __init__(self, stabilizer: Optional[Union[float, Stabilizer]] = 1e-6):
        super().__init__()
        if isinstance(stabilizer, (int, float)):
            # Use a more robust stabilizer for Flat rule
            stabilizer = Stabilizer(epsilon=stabilizer)
        self.stabilizer = stabilizer
        self.epsilon = stabilizer.epsilon if hasattr(stabilizer, 'epsilon') else 1e-6
    
    def forward(self, module: nn.Module, input: torch.Tensor, output: torch.Tensor) -> torch.Tensor:
        """Store the input for backward pass."""
        self.input = input[0] if isinstance(input, tuple) else input
        return output
    
    def backward(self, module: nn.Module, grad_input: tuple, grad_output: tuple) -> tuple:
        """
        Implement iNNvestigate's FlatRule backward pass logic.
        This matches the mathematical operations in iNNvestigate's explain_hook method.
        """
        if grad_output[0] is None:
            return grad_input
            
        relevance = grad_output[0]
        
        # Create flat weights (all ones) - matches iNNvestigate's FlatRule
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            flat_weight = torch.ones_like(module.weight)
            
            # Compute Zs: flat weights applied to ACTUAL input (not ones!)
            # This is the key fix - use actual input, not ones_input
            if isinstance(module, nn.Conv2d):
                zs = torch.nn.functional.conv2d(
                    self.input, flat_weight, None,
                    module.stride, module.padding, module.dilation, module.groups
                )
            else:  # Conv1d
                zs = torch.nn.functional.conv1d(
                    self.input, flat_weight, None,
                    module.stride, module.padding, module.dilation, module.groups
                )
                
        elif isinstance(module, nn.Linear):
            flat_weight = torch.ones_like(module.weight)
            # Use actual input, not ones_input
            zs = torch.nn.functional.linear(self.input, flat_weight, None)
        else:
            return grad_input
        
        # Apply enhanced SafeDivide operation with special handling for near-zero outputs
        # This is the CRITICAL FIX for the numerical instability issue
        zs_abs = torch.abs(zs)
        near_zero_threshold = self.epsilon * 1000  # More conservative threshold to prevent large values
        
        # Check if outputs are near zero (causing instability)
        near_zero_mask = zs_abs < near_zero_threshold
        
        if near_zero_mask.any():
            # For near-zero outputs, use a more conservative stabilization strategy
            # Use a larger threshold to keep ratios reasonable
            stabilized_near_zero = torch.where(
                zs >= 0,
                near_zero_threshold,  # Positive threshold for positive or zero values
                -near_zero_threshold  # Negative threshold for negative values
            )
            zs_stabilized = torch.where(
                near_zero_mask,
                stabilized_near_zero,
                self.stabilizer(zs)  # Use normal stabilization for non-zero outputs
            )
        else:
            zs_stabilized = self.stabilizer(zs)
            
        ratio = relevance / zs_stabilized
        
        # Additional safeguard: clip extreme values to prevent numerical overflow
        ratio = torch.clamp(ratio, min=-1e6, max=1e6)
        
        # Compute gradients with respect to input using flat weights
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            if isinstance(module, nn.Conv2d):
                grad_input_modified = torch.nn.functional.conv_transpose2d(
                    ratio, flat_weight, None,
                    module.stride, module.padding, 
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
            else:  # Conv1d
                grad_input_modified = torch.nn.functional.conv_transpose1d(
                    ratio, flat_weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
        elif isinstance(module, nn.Linear):
            # For linear: grad_input = ratio @ flat_weight
            grad_input_modified = torch.mm(ratio, flat_weight)
        else:
            grad_input_modified = grad_input[0]
        
        return (grad_input_modified,) + grad_input[1:]


class InnvestigateWSquareHook(Hook):
    """
    Custom WSquare hook that exactly matches iNNvestigate's WSquareRule implementation.
    
    From iNNvestigate: WSquareRule uses squared weights and no biases,
    then uses specific SafeDivide operations for relevance redistribution.
    """
    
    def __init__(self, stabilizer: Optional[Union[float, Stabilizer]] = 1e-6):
        super().__init__()
        if isinstance(stabilizer, (int, float)):
            stabilizer = Stabilizer(epsilon=stabilizer)
        self.stabilizer = stabilizer
    
    def forward(self, module: nn.Module, input: torch.Tensor, output: torch.Tensor) -> torch.Tensor:
        """Store the input and compute Zs for backward pass."""
        self.input = input[0] if isinstance(input, tuple) else input
        
        # Create squared weights - matches iNNvestigate's WSquareRule
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            squared_weight = module.weight ** 2
            
            # Compute Zs: forward pass with squared weights and ACTUAL input (not ones!)
            # This is the key fix - use actual input for proper WSquare computation
            if isinstance(module, nn.Conv2d):
                self.zs = torch.nn.functional.conv2d(
                    self.input, squared_weight, None,
                    module.stride, module.padding, module.dilation, module.groups
                )
            else:  # Conv1d
                self.zs = torch.nn.functional.conv1d(
                    self.input, squared_weight, None,
                    module.stride, module.padding, module.dilation, module.groups
                )
                
        elif isinstance(module, nn.Linear):
            squared_weight = module.weight ** 2
            # Use actual input for proper computation
            self.zs = torch.nn.functional.linear(self.input, squared_weight, None)
        else:
            self.zs = None
            
        return output
    
    def backward(self, module: nn.Module, grad_input: tuple, grad_output: tuple) -> tuple:
        """
        Implement iNNvestigate's WSquareRule backward pass logic.
        """
        if grad_output[0] is None:
            return grad_input
            
        relevance = grad_output[0]
        
        if self.zs is None:
            return grad_input
        
        # Apply SafeDivide operation (matches iNNvestigate)
        zs_stabilized = self.stabilizer(self.zs)
        ratio = relevance / zs_stabilized
        
        # Redistribute relevances using squared weights
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            squared_weight = module.weight ** 2
            
            if isinstance(module, nn.Conv2d):
                grad_input_modified = torch.nn.functional.conv_transpose2d(
                    ratio, squared_weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
            else:  # Conv1d
                grad_input_modified = torch.nn.functional.conv_transpose1d(
                    ratio, squared_weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
        elif isinstance(module, nn.Linear):
            squared_weight = module.weight ** 2
            grad_input_modified = torch.mm(ratio, squared_weight)
        else:
            grad_input_modified = grad_input[0]
        
        return (grad_input_modified,) + grad_input[1:]


class InnvestigateEpsilonHook(Hook):
    """
    Custom Epsilon hook that exactly matches iNNvestigate's EpsilonRule implementation.
    """
    
    def __init__(self, epsilon: float = 1e-6, bias: bool = True):
        super().__init__()
        self.epsilon = epsilon
        self.bias = bias
        self.stabilizer = Stabilizer(epsilon=epsilon)
    
    def forward(self, module: nn.Module, input: torch.Tensor, output: torch.Tensor) -> torch.Tensor:
        self.input = input[0] if isinstance(input, tuple) else input
        self.forward_output = output
        return output
    
    def backward(self, module: nn.Module, grad_input: tuple, grad_output: tuple) -> tuple:
        """
        Implement iNNvestigate's EpsilonRule backward pass with proper numerical stability.
        """
        if grad_output[0] is None:
            return grad_input
            
        relevance = grad_output[0]
        
        # Compute Zs = W * X + b (the denominator for relevance redistribution)
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            if isinstance(module, nn.Conv2d):
                zs = torch.nn.functional.conv2d(
                    self.input, module.weight, module.bias,
                    module.stride, module.padding, module.dilation, module.groups
                )
            else:  # Conv1d
                zs = torch.nn.functional.conv1d(
                    self.input, module.weight, module.bias,
                    module.stride, module.padding, module.dilation, module.groups
                )
        elif isinstance(module, nn.Linear):
            zs = torch.nn.functional.linear(self.input, module.weight, module.bias)
        else:
            return grad_input
        
        # Apply stabilization to Zs (this is the key to numerical stability)
        zs_stabilized = self.stabilizer(zs)
        ratio = relevance / zs_stabilized
        
        # Compute gradient redistribution using original weights
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            if isinstance(module, nn.Conv2d):
                grad_input_modified = torch.nn.functional.conv_transpose2d(
                    ratio, module.weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
            else:  # Conv1d
                grad_input_modified = torch.nn.functional.conv_transpose1d(
                    ratio, module.weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
        elif isinstance(module, nn.Linear):
            grad_input_modified = torch.mm(ratio, module.weight)
        else:
            grad_input_modified = grad_input[0]
        
        return (grad_input_modified,) + grad_input[1:]


# Custom composites using these hooks
def create_innvestigate_flat_composite():
    """Create a composite using InnvestigateFlatHook for all relevant layers."""
    
    # Use the same pattern as the working AdvancedLRPAnalyzer
    flat_hook = InnvestigateFlatHook()
    
    def module_map(ctx, name, module):
        if isinstance(module, (Convolution, Linear)):
            return flat_hook
        elif isinstance(module, (BatchNorm, Activation, AvgPool)):
            return None  # Pass-through for these layers
        return None
    
    return Composite(module_map=module_map)


def create_innvestigate_wsquare_composite():
    """Create a composite using InnvestigateWSquareHook for all relevant layers."""
    
    wsquare_hook = InnvestigateWSquareHook()
    
    def module_map(ctx, name, module):
        if isinstance(module, (Convolution, Linear)):
            return wsquare_hook
        elif isinstance(module, (BatchNorm, Activation, AvgPool)):
            return None  # Pass-through for these layers
        return None
    
    return Composite(module_map=module_map)


def create_innvestigate_epsilon_composite(epsilon: float = 1e-6):
    """Create a composite using InnvestigateEpsilonHook for all relevant layers."""
    
    epsilon_hook = InnvestigateEpsilonHook(epsilon=epsilon)
    
    def module_map(ctx, name, module):
        if isinstance(module, (Convolution, Linear)):
            return epsilon_hook
        elif isinstance(module, (BatchNorm, Activation, AvgPool)):
            return None  # Pass-through for these layers
        return None
    
    return Composite(module_map=module_map)


class InnvestigateZPlusHook(Hook):
    """
    Custom ZPlus hook that exactly matches iNNvestigate's ZPlusRule implementation.
    
    From iNNvestigate: ZPlusRule uses only positive weights and no biases,
    effectively implementing LRP-0 rule for positive contributions only.
    """
    
    def __init__(self, stabilizer: Optional[Union[float, Stabilizer]] = 1e-6):
        super().__init__()
        if isinstance(stabilizer, (int, float)):
            stabilizer = Stabilizer(epsilon=stabilizer)
        self.stabilizer = stabilizer
    
    def forward(self, module: nn.Module, input: torch.Tensor, output: torch.Tensor) -> torch.Tensor:
        self.input = input[0] if isinstance(input, tuple) else input
        return output
    
    def backward(self, module: nn.Module, grad_input: tuple, grad_output: tuple) -> tuple:
        """
        Implement iNNvestigate's ZPlusRule backward pass logic.
        """
        if grad_output[0] is None:
            return grad_input
            
        relevance = grad_output[0]
        
        # Use only positive weights - matches iNNvestigate's ZPlusRule
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            positive_weight = torch.clamp(module.weight, min=0)
            
            # Compute forward pass with positive weights only
            if isinstance(module, nn.Conv2d):
                zs = torch.nn.functional.conv2d(
                    self.input, positive_weight, None,
                    module.stride, module.padding, module.dilation, module.groups
                )
            else:  # Conv1d
                zs = torch.nn.functional.conv1d(
                    self.input, positive_weight, None,
                    module.stride, module.padding, module.dilation, module.groups
                )
                
        elif isinstance(module, nn.Linear):
            positive_weight = torch.clamp(module.weight, min=0)
            zs = torch.nn.functional.linear(self.input, positive_weight, None)
        else:
            return grad_input
        
        # Apply stabilization
        zs_stabilized = self.stabilizer(zs)
        ratio = relevance / zs_stabilized
        
        # Compute gradients using positive weights
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            if isinstance(module, nn.Conv2d):
                grad_input_modified = torch.nn.functional.conv_transpose2d(
                    ratio, positive_weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
            else:  # Conv1d
                grad_input_modified = torch.nn.functional.conv_transpose1d(
                    ratio, positive_weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
        elif isinstance(module, nn.Linear):
            grad_input_modified = torch.mm(ratio, positive_weight)
        else:
            grad_input_modified = grad_input[0]
        
        return (grad_input_modified,) + grad_input[1:]


class InnvestigateAlphaBetaHook(Hook):
    """
    Custom AlphaBeta hook that exactly matches iNNvestigate's AlphaBetaRule implementation.
    
    From iNNvestigate: AlphaBetaRule separates positive and negative contributions
    and weights them differently using alpha and beta parameters.
    """
    
    def __init__(self, alpha: float = 2.0, beta: float = 1.0, stabilizer: Optional[Union[float, Stabilizer]] = 1e-6):
        super().__init__()
        self.alpha = alpha
        self.beta = beta
        if isinstance(stabilizer, (int, float)):
            stabilizer = Stabilizer(epsilon=stabilizer)
        self.stabilizer = stabilizer
    
    def forward(self, module: nn.Module, input: torch.Tensor, output: torch.Tensor) -> torch.Tensor:
        self.input = input[0] if isinstance(input, tuple) else input
        return output
    
    def backward(self, module: nn.Module, grad_input: tuple, grad_output: tuple) -> tuple:
        """
        Implement iNNvestigate's AlphaBetaRule backward pass logic.
        """
        if grad_output[0] is None:
            return grad_input
            
        relevance = grad_output[0]
        
        # Separate positive and negative weights
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            positive_weight = torch.clamp(module.weight, min=0)
            negative_weight = torch.clamp(module.weight, max=0)
            
            # Compute forward passes for positive and negative parts
            if isinstance(module, nn.Conv2d):
                zs_pos = torch.nn.functional.conv2d(
                    self.input, positive_weight, None,
                    module.stride, module.padding, module.dilation, module.groups
                )
                zs_neg = torch.nn.functional.conv2d(
                    self.input, negative_weight, None,
                    module.stride, module.padding, module.dilation, module.groups
                )
            else:  # Conv1d
                zs_pos = torch.nn.functional.conv1d(
                    self.input, positive_weight, None,
                    module.stride, module.padding, module.dilation, module.groups
                )
                zs_neg = torch.nn.functional.conv1d(
                    self.input, negative_weight, None,
                    module.stride, module.padding, module.dilation, module.groups
                )
                
        elif isinstance(module, nn.Linear):
            positive_weight = torch.clamp(module.weight, min=0)
            negative_weight = torch.clamp(module.weight, max=0)
            zs_pos = torch.nn.functional.linear(self.input, positive_weight, None)
            zs_neg = torch.nn.functional.linear(self.input, negative_weight, None)
        else:
            return grad_input
        
        # Apply alpha-beta weighting
        zs_combined = self.alpha * zs_pos + self.beta * zs_neg
        zs_stabilized = self.stabilizer(zs_combined)
        ratio = relevance / zs_stabilized
        
        # Compute gradients using alpha-beta weighted combination
        combined_weight = self.alpha * positive_weight + self.beta * negative_weight
        
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            if isinstance(module, nn.Conv2d):
                grad_input_modified = torch.nn.functional.conv_transpose2d(
                    ratio, combined_weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
            else:  # Conv1d
                grad_input_modified = torch.nn.functional.conv_transpose1d(
                    ratio, combined_weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
        elif isinstance(module, nn.Linear):
            grad_input_modified = torch.mm(ratio, combined_weight)
        else:
            grad_input_modified = grad_input[0]
        
        return (grad_input_modified,) + grad_input[1:]


class InnvestigateGammaHook(Hook):
    """
    Custom Gamma hook that exactly matches iNNvestigate's GammaRule implementation.
    
    From iNNvestigate: GammaRule modifies weights by adding a small gamma value
    to increase stability and handle edge cases.
    """
    
    def __init__(self, gamma: float = 0.25, stabilizer: Optional[Union[float, Stabilizer]] = 1e-6):
        super().__init__()
        self.gamma = gamma
        if isinstance(stabilizer, (int, float)):
            stabilizer = Stabilizer(epsilon=stabilizer)
        self.stabilizer = stabilizer
    
    def forward(self, module: nn.Module, input: torch.Tensor, output: torch.Tensor) -> torch.Tensor:
        self.input = input[0] if isinstance(input, tuple) else input
        return output
    
    def backward(self, module: nn.Module, grad_input: tuple, grad_output: tuple) -> tuple:
        """
        Implement iNNvestigate's GammaRule backward pass logic.
        """
        if grad_output[0] is None:
            return grad_input
            
        relevance = grad_output[0]
        
        # Apply gamma modification to weights
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            gamma_weight = module.weight + self.gamma
            
            # Compute forward pass with gamma-modified weights
            if isinstance(module, nn.Conv2d):
                zs = torch.nn.functional.conv2d(
                    self.input, gamma_weight, None,
                    module.stride, module.padding, module.dilation, module.groups
                )
            else:  # Conv1d
                zs = torch.nn.functional.conv1d(
                    self.input, gamma_weight, None,
                    module.stride, module.padding, module.dilation, module.groups
                )
                
        elif isinstance(module, nn.Linear):
            gamma_weight = module.weight + self.gamma
            zs = torch.nn.functional.linear(self.input, gamma_weight, None)
        else:
            return grad_input
        
        # Apply stabilization
        zs_stabilized = self.stabilizer(zs)
        ratio = relevance / zs_stabilized
        
        # Compute gradients using gamma-modified weights
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            if isinstance(module, nn.Conv2d):
                grad_input_modified = torch.nn.functional.conv_transpose2d(
                    ratio, gamma_weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
            else:  # Conv1d
                grad_input_modified = torch.nn.functional.conv_transpose1d(
                    ratio, gamma_weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
        elif isinstance(module, nn.Linear):
            grad_input_modified = torch.mm(ratio, gamma_weight)
        else:
            grad_input_modified = grad_input[0]
        
        return (grad_input_modified,) + grad_input[1:]


class InnvestigateZBoxHook(Hook):
    """
    Custom ZBox hook that exactly matches iNNvestigate's ZBoxRule implementation.
    
    From iNNvestigate: ZBoxRule applies input bounds constraints during
    relevance propagation to handle edge cases at input boundaries.
    """
    
    def __init__(self, low: float = 0.0, high: float = 1.0, stabilizer: Optional[Union[float, Stabilizer]] = 1e-6):
        super().__init__()
        self.low = low
        self.high = high
        if isinstance(stabilizer, (int, float)):
            stabilizer = Stabilizer(epsilon=stabilizer)
        self.stabilizer = stabilizer
    
    def forward(self, module: nn.Module, input: torch.Tensor, output: torch.Tensor) -> torch.Tensor:
        self.input = input[0] if isinstance(input, tuple) else input
        return output
    
    def backward(self, module: nn.Module, grad_input: tuple, grad_output: tuple) -> tuple:
        """
        Implement iNNvestigate's ZBoxRule backward pass logic.
        """
        if grad_output[0] is None:
            return grad_input
            
        relevance = grad_output[0]
        
        # Create bounded inputs for upper and lower bounds
        input_low = torch.full_like(self.input, self.low)
        input_high = torch.full_like(self.input, self.high)
        
        # Compute forward passes for different input bounds
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            if isinstance(module, nn.Conv2d):
                # Standard forward pass
                zs = torch.nn.functional.conv2d(
                    self.input, module.weight, module.bias,
                    module.stride, module.padding, module.dilation, module.groups
                )
                # Forward pass with low bound
                zs_low = torch.nn.functional.conv2d(
                    input_low, module.weight, module.bias,
                    module.stride, module.padding, module.dilation, module.groups
                )
                # Forward pass with high bound  
                zs_high = torch.nn.functional.conv2d(
                    input_high, module.weight, module.bias,
                    module.stride, module.padding, module.dilation, module.groups
                )
            else:  # Conv1d
                zs = torch.nn.functional.conv1d(
                    self.input, module.weight, module.bias,
                    module.stride, module.padding, module.dilation, module.groups
                )
                zs_low = torch.nn.functional.conv1d(
                    input_low, module.weight, module.bias,
                    module.stride, module.padding, module.dilation, module.groups
                )
                zs_high = torch.nn.functional.conv1d(
                    input_high, module.weight, module.bias,
                    module.stride, module.padding, module.dilation, module.groups
                )
                
        elif isinstance(module, nn.Linear):
            zs = torch.nn.functional.linear(self.input, module.weight, module.bias)
            zs_low = torch.nn.functional.linear(input_low, module.weight, module.bias)
            zs_high = torch.nn.functional.linear(input_high, module.weight, module.bias)
        else:
            return grad_input
        
        # Apply ZBox logic - use bounds to constrain relevance flow
        zs_diff = zs - zs_low - zs_high
        zs_stabilized = self.stabilizer(zs_diff)
        ratio = relevance / zs_stabilized
        
        # Compute gradients with bounded constraint
        input_diff = self.input - input_low - input_high
        if isinstance(module, (nn.Conv2d, nn.Conv1d)):
            if isinstance(module, nn.Conv2d):
                grad_input_modified = torch.nn.functional.conv_transpose2d(
                    ratio, module.weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
            else:  # Conv1d
                grad_input_modified = torch.nn.functional.conv_transpose1d(
                    ratio, module.weight, None,
                    module.stride, module.padding,
                    output_padding=0, groups=module.groups, dilation=module.dilation
                )
        elif isinstance(module, nn.Linear):
            grad_input_modified = torch.mm(ratio, module.weight)
        else:
            grad_input_modified = grad_input[0]
        
        # Apply input bounds constraint
        grad_input_modified = grad_input_modified * input_diff
        
        return (grad_input_modified,) + grad_input[1:]


# Additional composite creators for the new hooks
def create_innvestigate_zplus_composite():
    """Create a composite using InnvestigateZPlusHook for all relevant layers."""
    
    zplus_hook = InnvestigateZPlusHook()
    
    def module_map(ctx, name, module):
        if isinstance(module, (Convolution, Linear)):
            return zplus_hook
        elif isinstance(module, (BatchNorm, Activation, AvgPool)):
            return None  # Pass-through for these layers
        return None
    
    return Composite(module_map=module_map)


def create_innvestigate_alphabeta_composite(alpha: float = 2.0, beta: float = 1.0):
    """Create a composite using InnvestigateAlphaBetaHook for all relevant layers."""
    
    alphabeta_hook = InnvestigateAlphaBetaHook(alpha=alpha, beta=beta)
    
    def module_map(ctx, name, module):
        if isinstance(module, (Convolution, Linear)):
            return alphabeta_hook
        elif isinstance(module, (BatchNorm, Activation, AvgPool)):
            return None  # Pass-through for these layers
        return None
    
    return Composite(module_map=module_map)


def create_innvestigate_gamma_composite(gamma: float = 0.25):
    """Create a composite using InnvestigateGammaHook for all relevant layers."""
    
    gamma_hook = InnvestigateGammaHook(gamma=gamma)
    
    def module_map(ctx, name, module):
        if isinstance(module, (Convolution, Linear)):
            return gamma_hook
        elif isinstance(module, (BatchNorm, Activation, AvgPool)):
            return None  # Pass-through for these layers
        return None
    
    return Composite(module_map=module_map)


def create_innvestigate_zbox_composite(low: float = 0.0, high: float = 1.0):
    """Create a composite using InnvestigateZBoxHook for all relevant layers."""
    
    zbox_hook = InnvestigateZBoxHook(low=low, high=high)
    
    def module_map(ctx, name, module):
        if isinstance(module, (Convolution, Linear)):
            return zbox_hook
        elif isinstance(module, (BatchNorm, Activation, AvgPool)):
            return None  # Pass-through for these layers
        return None
    
    return Composite(module_map=module_map)


def create_innvestigate_sequential_composite(first_rule: str = "zbox", middle_rule: str = "alphabeta", 
                                           last_rule: str = "epsilon", first_layer_name: str = None, 
                                           last_layer_name: str = None, **kwargs):
    """
    Create a sequential composite that uses different iNNvestigate-compatible hooks
    for different layers, matching iNNvestigate's sequential rule application.
    
    Args:
        first_rule: Rule to apply to first layers (default: "zbox")
        middle_rule: Rule to apply to middle layers (default: "alphabeta") 
        last_rule: Rule to apply to last layers (default: "epsilon")
        first_layer_name: Name pattern for first layers
        last_layer_name: Name pattern for last layers
        **kwargs: Additional parameters for rule creation
    """
    
    # Create hooks for each rule type
    if first_rule == "zbox":
        first_hook = InnvestigateZBoxHook(low=kwargs.get("low", 0.0), high=kwargs.get("high", 1.0))
    elif first_rule == "epsilon":
        first_hook = InnvestigateEpsilonHook(epsilon=kwargs.get("epsilon", 1e-6))
    elif first_rule == "flat":
        first_hook = InnvestigateFlatHook()
    elif first_rule == "wsquare":
        first_hook = InnvestigateWSquareHook()
    else:
        first_hook = InnvestigateEpsilonHook(epsilon=kwargs.get("epsilon", 1e-6))
    
    if middle_rule == "alphabeta":
        middle_hook = InnvestigateAlphaBetaHook(
            alpha=kwargs.get("alpha", 1.0), 
            beta=kwargs.get("beta", 0.0)
        )
    elif middle_rule == "zplus":
        middle_hook = InnvestigateZPlusHook()
    elif middle_rule == "epsilon":
        middle_hook = InnvestigateEpsilonHook(epsilon=kwargs.get("epsilon", 1e-6))
    elif middle_rule == "flat":
        middle_hook = InnvestigateFlatHook()
    elif middle_rule == "wsquare":
        middle_hook = InnvestigateWSquareHook()
    elif middle_rule == "gamma":
        middle_hook = InnvestigateGammaHook(gamma=kwargs.get("gamma", 0.25))
    else:
        middle_hook = InnvestigateAlphaBetaHook(alpha=1.0, beta=0.0)
    
    if last_rule == "epsilon":
        last_hook = InnvestigateEpsilonHook(epsilon=kwargs.get("epsilon", 1e-6))
    elif last_rule == "flat":
        last_hook = InnvestigateFlatHook()
    elif last_rule == "wsquare":
        last_hook = InnvestigateWSquareHook()
    elif last_rule == "alphabeta":
        last_hook = InnvestigateAlphaBetaHook(
            alpha=kwargs.get("alpha", 2.0), 
            beta=kwargs.get("beta", 1.0)
        )
    elif last_rule == "zplus":
        last_hook = InnvestigateZPlusHook()
    elif last_rule == "gamma":
        last_hook = InnvestigateGammaHook(gamma=kwargs.get("gamma", 0.25))
    else:
        last_hook = InnvestigateEpsilonHook(epsilon=kwargs.get("epsilon", 1e-6))
    
    def module_map(ctx, name, module):
        if isinstance(module, (Convolution, Linear)):
            # Apply rules based on layer position/name
            if first_layer_name and first_layer_name in name:
                return first_hook
            elif last_layer_name and last_layer_name in name:
                return last_hook
            else:
                # Default to middle rule for most layers
                return middle_hook
        elif isinstance(module, (BatchNorm, Activation, AvgPool)):
            return None  # Pass-through for these layers
        return None
    
    return Composite(module_map=module_map)